// LuxSpectralis.c
// RGB LED color wash
// written by Dale Wheat - 2 December 2008

// device = ATtiny13
// clock = internal 9.6 MHz oscillator, longest startup time
// max ISP voltage 4.5 VDC - sometimes needs to be less (3.0 V)
// BOD = 1.8V

#include <avr/io.h>
#include <avr/interrupt.h>

// These registers are available on the ATtiny13A but not the original ATtiny13

// Brown out detector control register

#define BODCR _SFR_IO8(0x30)
#define BODSE 0
#define BODS 1

// Power reduction register

#define PRR _SFR_IO8(0x3C)
#define PRADC 0
#define PRTIM0 1

///////////////////////////////////////////////////////////////////////////////
// macros
///////////////////////////////////////////////////////////////////////////////

#define nop() asm("nop")

#define sbi(port, bit) (port) |= (1 << (bit))
#define cbi(port, bit) (port) &= ~(1 << (bit))

typedef enum {
	MODE_POWER_ON = 0, // power on behavior
	MODE_WHITE_5MIN = 0,
	MODE_WHITE_30MIN,
	MODE_WHITE_25,
	MODE_WHITE_50,
	MODE_WHITE_100,
	MODE_RED_25,
	MODE_RED_100,
	MODE_GREEN_25,
	MODE_GREEN_100,
	MODE_BLUE_25,
	MODE_BLUE_100,
	MODE_YELLOW_25,
	MODE_YELLOW_100,
	MODE_AQUA_25,
	MODE_AQUA_100,
	MODE_VIOLET_25,
	MODE_VIOLET_100,
	MODE_WASH1_25,
	MODE_WASH1_100,
	MODE_WASH2_25,
	MODE_WASH2_100,
	MODE_WASH3_25,
	MODE_WASH3_100,
	MODE_WASH4_25,
	MODE_WASH4_100,
	MODE_FLASH_WHITE,
	MODE_FLASH_RED,
	MODE_FLASH_GREEN,
	MODE_FLASH_BLUE,
	MODE_FLASH_RED_GREEN,
	MODE_FLASH_RED_BLUE,
	MODE_FLASH_GREEN_BLUE,
	MODE_BLINK_WHITE,
	MODE_BLINK_RED,
	MODE_BLINK_GREEN,
	MODE_BLINK_BLUE,
	MODE_MAX // off
} MODE;

volatile MODE mode __attribute__ ((section (".noinit")));

///////////////////////////////////////////////////////////////////////////////
// init() - initialize everything
// note:  this "function" is in section .init3 so it is executed before main()
///////////////////////////////////////////////////////////////////////////////

void init(void) __attribute__ ((naked, section(".init3")));
void init(void) {

	// turn off unused peripherals to save power

	PRR = 0<<PRTIM0 | 1<<PRADC; // power down ADC
	ACSR = 1<<ACD; // disable analog comparator
	DIDR0 = 1<<ADC3D | 1<<ADC2D | 1<<ADC1D | 1<<ADC0D | 1<<AIN1D | 1<<AIN0D; // disable all digital inputs

	// determine cause of device reset;  act accordingly

	if(bit_is_set(MCUSR, PORF)) {
		mode = MODE_POWER_ON; // power on!
		//mode = MODE_BLINK_WHITE; // *** debug ***
	} else if(bit_is_set(MCUSR, EXTRF)) {
		mode++; // advance mode
		if(mode > MODE_MAX) {
			mode = MODE_POWER_ON; // reset mode
		}
	}

	MCUSR = 0; // reset bits

	// initialize ATtiny13 input & output port

	// PORTB

	//	PB0		5		MOSI/AIN0/OC0A/PCINT0		common cathode driver PWM output for lighting LEDs
	//	PB1		6		MISO/AIN1/OC0B/INT0/PCINT1	common cathode high side driver output for reverse biasing LEDs
	//	PB2		7		SCK/ADC1/T0/PCINT2			red LED input/output
	//	PB3		2		PCINT3/CLKI/ADC3			green LED input/output
	//	PB4		3		PCINT4/ADC2					blue LED input/output
	//	PB5		1		PCINT5/-RESET/ADC0/dW		MODE advance pushbutton aka RESET button

	PORTB = 1<<PORTB5 | 0<<PORTB4 | 0<<PORTB3 | 0<<PORTB2 | 1<<PORTB1 | 0<<PORTB0;
	DDRB = 0<<DDB5 | 1<<DDB4 | 1<<DDB3 | 1<<DDB2 | 1<<DDB1 | 1<<DDB0;

	// initialize ATtiny13 timer/counter

	TCCR0A = 1<<COM0A1 | 0<<COM0A0 | 0<<COM0B1 | 0<<COM0B0 | 1<<WGM01 | 1<<WGM00;
	TCCR0B = 0<<FOC0A | 0<<FOC0B | 0<<WGM02 | 0<<CS02 | 0<<CS01 | 1<<CS00;
	TIMSK0 = 0<<OCIE0B | 0<<OCIE0A | 1<<TOIE0; // interrupts

	sei(); // enable global interrupts
}

///////////////////////////////////////////////////////////////////////////////
// timing & delay functions
///////////////////////////////////////////////////////////////////////////////

void delay(unsigned char n) {

	while(n) {
		MCUCR = 1<<PUD | 1<<SE | 0<<SM1 | 0<<SM0 | 0<<ISC01 | 0<<ISC00; // idle mode
		asm("sleep"); // go to sleep to save power; wait for timer interrupt
		n--;
	}
}

///////////////////////////////////////////////////////////////////////////////
// pseudorandom number generator
///////////////////////////////////////////////////////////////////////////////

unsigned int prand(void) {

    static unsigned int prand_value = 0xDA1E; // randomly seeded ;)
    
    prand_value = (prand_value >> 1) ^ (-(prand_value & 1) & 0xd001);

    return prand_value;
}

///////////////////////////////////////////////////////////////////////////////
// timed_light() - bring up white light for period then dim over same period
///////////////////////////////////////////////////////////////////////////////

void timed_light(unsigned char period) {

	unsigned long int i; // misc. iterator

	PORTB = 1<<PB4 | 1<<PB3 | 1<<PB2 | 1<<PB1; // white
	for(OCR0A = 0; OCR0A < 255; OCR0A++) delay(period); // ramp up
	for(i = 512 * period; i > 0; i--) delay(255); // full bright for specified period

	for(OCR0A = 255; OCR0A > 0; OCR0A--) {
		for(i = 2 * period; i > 0; i--) delay(255); // dim over specified period
	}
	PORTB = 0<<PB4 | 0<<PB3 | 0<<PB2 | 1<<PB1; // black
	OCR0A = 0; // now it's off
	mode = MODE_MAX; // now go into deep sleep mode
}

///////////////////////////////////////////////////////////////////////////////
// color_wash() - randomly change colors
///////////////////////////////////////////////////////////////////////////////

void color_wash(unsigned char pace) {

	unsigned char red = 0, green = 0, blue = 0; // intensities
	unsigned char target_red = 0, target_green = 0, target_blue = 0; // target values
	unsigned char pwm; // PWM iterator
	unsigned char i; // loop iterator

	while(1) {

		pwm = 0;

		for(i = 0; i < pace; i++) {

			do {
				if(red < pwm) {
					sbi(PORTB, PB2); // red is on
				} else {
					cbi(PORTB, PB2); // red is off
				}

				if(green < pwm) {
					sbi(PORTB, PB3); // green is on
				} else {
					cbi(PORTB, PB3); // green is off
				}

				if(blue < pwm) {
					sbi(PORTB, PB4); // blue is on
				} else {
					cbi(PORTB, PB4); // blue is off
				}

				pwm++;

				delay(1);

			} while(pwm > 0);
		}

		if(red > target_red) {
			red--;
		} else if(red < target_red) {
			red++;
		} else {
			target_red = prand(); // new target
		}

		if(green > target_green) {
			green--;
		} else if(green < target_green) {
			green++;
		} else {
			target_green = prand(); // new target
		}

		if(blue > target_blue) {
			blue--;
		} else if(blue < target_blue) {
			blue++;
		} else {
			target_blue = prand(); // new target
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// flash() - alternate between two colors, repeat
///////////////////////////////////////////////////////////////////////////////

void flash(unsigned char color1, unsigned char time1, unsigned char color2, unsigned char time2) {

	unsigned char prescaler;

	OCR0A = 0xFF; // 100% brightness

	while(1) {

		PORTB = color1 | 1<<PB1;
		for(prescaler = time1; prescaler > 0; prescaler--) delay(255);

		PORTB = color2 | 1<<PB1;
		for(prescaler = time2; prescaler > 0; prescaler--) delay(255);
	}
}

///////////////////////////////////////////////////////////////////////////////
// main() - main program function
///////////////////////////////////////////////////////////////////////////////

void main(void) {

	switch(mode) {

		case MODE_WHITE_5MIN: // ramp up to full white for 5 mintutes then fade over 5 minutes
			timed_light(42); // 5 minutes
			break;

		case MODE_WHITE_30MIN: // ramp up (a little slower) to full white for 30 mintutes then fade over 30 minutes
			timed_light(255); // 30 minutes
			break;

		case MODE_WHITE_25:
			PORTB = 1<<PB4 | 1<<PB3 | 1<<PB2 | 1<<PB1; // white
			OCR0A = 0x40; // 25% brightness
			break;

		case MODE_WHITE_50:
			PORTB = 1<<PB4 | 1<<PB3 | 1<<PB2 | 1<<PB1; // white
			OCR0A = 0x80; // 50% brightness
			break;

		case MODE_WHITE_100:
			PORTB = 1<<PB4 | 1<<PB3 | 1<<PB2 | 1<<PB1; // white
			OCR0A = 0xFF; // 100% brightness
			break;

		case MODE_RED_25:
			PORTB = 0<<PB4 | 0<<PB3 | 1<<PB2 | 1<<PB1; // red
			OCR0A = 0x40; // 25% brightness
			break;

		case MODE_RED_100:
			PORTB = 0<<PB4 | 0<<PB3 | 1<<PB2 | 1<<PB1; // red
			OCR0A = 0xFF; // 100% brightness
			break;

		case MODE_GREEN_25:
			PORTB = 0<<PB4 | 1<<PB3 | 0<<PB2 | 1<<PB1; // green
			OCR0A = 0x40; // 25% brightness
			break;

		case MODE_GREEN_100:
			PORTB = 0<<PB4 | 1<<PB3 | 0<<PB2 | 1<<PB1; // green
			OCR0A = 0xFF; // 100% brightness
			break;

		case MODE_BLUE_25:
			PORTB = 1<<PB4 | 0<<PB3 | 0<<PB2 | 1<<PB1; // blue
			OCR0A = 0x40; // 25% brightness
			break;

		case MODE_BLUE_100:
			PORTB = 1<<PB4 | 0<<PB3 | 0<<PB2 | 1<<PB1; // blue
			OCR0A = 0xFF; // 100% brightness
			break;

		case MODE_YELLOW_25:
			PORTB = 0<<PB4 | 1<<PB3 | 1<<PB2 | 1<<PB1; // yellow
			OCR0A = 0x40; // 25% brightness
			break;

		case MODE_YELLOW_100:
			PORTB = 0<<PB4 | 1<<PB3 | 1<<PB2 | 1<<PB1; // yellow
			OCR0A = 0xFF; // 100% brightness
			break;

		case MODE_AQUA_25:
			PORTB = 1<<PB4 | 1<<PB3 | 0<<PB2 | 1<<PB1; // aqua
			OCR0A = 0x40; // 25% brightness
			break;

		case MODE_AQUA_100:
			PORTB = 1<<PB4 | 1<<PB3 | 0<<PB2 | 1<<PB1; // aqua
			OCR0A = 0xFF; // 100% brightness
			break;

		case MODE_VIOLET_25:
			PORTB = 1<<PB4 | 0<<PB3 | 1<<PB2 | 1<<PB1; // violet
			OCR0A = 0x40; // 25% brightness
			break;

		case MODE_VIOLET_100:
			PORTB = 1<<PB4 | 0<<PB3 | 1<<PB2 | 1<<PB1; // violet
			OCR0A = 0xFF; // 100% brightness
			break;

		case MODE_WASH1_25:
			OCR0A = 0x40; // 25% brightness, fast
			color_wash(1);
			break;

		case MODE_WASH1_100:
			OCR0A = 0xFF; // 100% brightness, fast
			color_wash(1);
			break;

		case MODE_WASH2_25:
			OCR0A = 0x40; // 25% brightness, moderate
			color_wash(3);
			break;

		case MODE_WASH2_100:
			OCR0A = 0xFF; // 100% brightness, moderate
			color_wash(3);
			break;

		case MODE_WASH3_25:
			OCR0A = 0x40; // 25% brightness, slow
			color_wash(10);
			break;

		case MODE_WASH3_100:
			OCR0A = 0xFF; // 100% brightness, slow
			color_wash(10);
			break;

		case MODE_WASH4_25:
			OCR0A = 0x40; // 25% brightness, really slow
			color_wash(100);
			break;

		case MODE_WASH4_100:
			OCR0A = 0xFF; // 100% brightness, really slow
			color_wash(100);
			break;

		case MODE_FLASH_WHITE:
			flash(1<<PB4 | 1<<PB3 | 1<<PB2, 73, 0, 73); // flash white
			break;

		case MODE_FLASH_RED:
			flash(0<<PB4 | 0<<PB3 | 1<<PB2, 73, 0, 73); // flash red
			break;

		case MODE_FLASH_GREEN:
			flash(0<<PB4 | 1<<PB3 | 0<<PB2, 73, 0, 73); // flash green
			break;

		case MODE_FLASH_BLUE:
			flash(1<<PB4 | 0<<PB3 | 0<<PB2, 73, 0, 73); // flash blue
			break;

		case MODE_FLASH_RED_GREEN:
			flash(1<<PB2, 73, 1<<PB3, 73); // flash red & green
			break;

		case MODE_FLASH_RED_BLUE:
			flash(1<<PB2, 73, 1<<PB4, 73); // flash red & blue
			break;

		case MODE_FLASH_GREEN_BLUE:
			flash(1<<PB3, 73, 1<<PB4, 73); // flash green & blue
			break;

		case MODE_BLINK_WHITE:
			flash(1<<PB4 | 1<<PB3 | 1<<PB2, 2, 0, 144); // blink white
			break;

		case MODE_BLINK_RED:
			flash(0<<PB4 | 0<<PB3 | 1<<PB2, 5, 0, 141); // blink red
			break;

		case MODE_BLINK_GREEN:
			flash(0<<PB4 | 1<<PB3 | 0<<PB2, 5, 0, 141); // blink green
			break;

		case MODE_BLINK_BLUE:
			flash(1<<PB4 | 0<<PB3 | 0<<PB2, 5, 0, 141); // blink blue
			break;

		case MODE_MAX: // this is the last mode;  go into low power mode
			// deepest sleep mode
			cli(); // disable interrupts
			PORTB = 0b00000010; // all LEDs off, not reverse biased either
			PRR = 1<<PRTIM0 | 1<<PRADC; // power down timer/counter0 & ADC
			BODCR = 1<<BODS | 1<<BODSE; // enable BOD disable during sleep, step 1
			BODCR = 1<<BODS | 0<<BODSE; // step 2
			MCUCR = 1<<PUD | 1<<SE | 1<<SM1 | 0<<SM0 | 0<<ISC01 | 0<<ISC00; // select "power down" mode
			asm("sleep"); // go to sleep to save power
			break;
	}
}

///////////////////////////////////////////////////////////////////////////////
// timer/counter0 overflow interrupt handler
///////////////////////////////////////////////////////////////////////////////

ISR(TIM0_OVF_vect) __attribute__ ((naked));
ISR(TIM0_OVF_vect) {

	asm("reti"); // return from interrupt
}

///////////////////////////////////////////////////////////////////////////////

// [end-of-file]
